/*  

    Panda -  A modular programming language.
    written by Giovanni Petrantoni.
    Copyright (c) 2003, Giovanni Petrantoni.
    http://xpanda.sf.net/
    
    moduleloader.cpp - main Panda classes
    
    Panda is under the terms of the Artistic License :
    
    The Artistic License
    
    Preamble

    The intent of this document is to state the conditions under which a Package may be copied, 
    such that the Copyright Holder maintains some semblance of artistic control over the development of the package, 
    while giving the users of the package the right to use and distribute the Package in a more-or-less customary fashion,
    plus the right to make reasonable modifications.

    Definitions:

    "Package" refers to the collection of files distributed by the Copyright Holder, 
    and derivatives of that collection of files created through textual modification.

    "Standard Version" refers to such a Package if it has not been modified, 
    or has been modified in accordance with the wishes of the Copyright Holder.

    "Copyright Holder" is whoever is named in the copyright or copyrights for the package.

    "You" is you, if you're thinking about copying or distributing this Package.

    "Reasonable copying fee" is whatever you can justify on the basis of media cost, duplication charges, 
    time of people involved, and so on. (You will not be required to justify it to the Copyright Holder, 
    but only to the computing community at large as a market that must bear the fee.)

    "Freely Available" means that no fee is charged for the item itself, 
    though there may be fees involved in handling the item. 
    It also means that recipients of the item may redistribute it under the same conditions they received it.

    1. You may make and give away verbatim copies of the source form of the Standard Version of this Package without restriction, 
    provided that you duplicate all of the original copyright notices and associated disclaimers.

    2. You may apply bug fixes, portability fixes and other modifications derived from the Public Domain or from the Copyright Holder. 
    A Package modified in such a way shall still be considered the Standard Version.

    3. You may otherwise modify your copy of this Package in any way, 
    provided that you insert a prominent notice in each changed file stating how and when you changed that file, 
    and provided that you do at least ONE of the following:

        a) place your modifications in the Public Domain or otherwise make them Freely Available, 
        such as by posting said modifications to Usenet or an equivalent medium, 
        or placing the modifications on a major archive site such as ftp.uu.net, 
        or by allowing the Copyright Holder to include your modifications in the Standard Version of the Package.

        b) use the modified Package only within your corporation or organization.

        c) rename any non-standard executables so the names do not conflict with standard executables, 
        which must also be provided, and provide a separate manual page for each non-standard executable 
        that clearly documents how it differs from the Standard Version.

        d) make other distribution arrangements with the Copyright Holder.

    4. You may distribute the programs of this Package in object code or executable form, 
    provided that you do at least ONE of the following:

    a) distribute a Standard Version of the executables and library files, 
    together with instructions (in the manual page or equivalent) on where to get the Standard Version.

    b) accompany the distribution with the machine-readable source of the Package with your modifications.

    c) accompany any non-standard executables with their corresponding Standard Version executables, 
    giving the non-standard executables non-standard names, and clearly documenting the differences in manual pages (or equivalent), 
    together with instructions on where to get the Standard Version.

    d) make other distribution arrangements with the Copyright Holder.

    5. You may charge a reasonable copying fee for any distribution of this Package. 
    You may charge any fee you choose for support of this Package. You may not charge a fee for this Package itself. 
    However, you may distribute this Package in aggregate with other (possibly commercial) programs as part of 
    a larger (possibly commercial) software distribution provided that you do not advertise this Package as a product of your own.

    6. The scripts and library files supplied as input to or produced as output from the programs of this Package do not automatically 
    fall under the copyright of this Package, but belong to whomever generated them, and may be sold commercially, and may be aggregated with this Package.

    7. C or perl subroutines supplied by you and linked into this Package shall not be considered part of this Package.

    8. The name of the Copyright Holder may not be used to endorse or promote products derived from this software without specific prior written permission.

    9. THIS PACKAGE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, 
    INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.

    The End.

*/

#include "moduleloader.hpp"
#include <fcntl.h>

#define CODE_1 'jJOh'
#define CODE_2 '0173'
#define CODE_3 'FiAL'
#define CODE_4 '2F28'

#define BINCREATE O_WRONLY | O_CREAT | O_TRUNC

list<ModuleLoader*> ModuleLoader::Audio_StartPoints;
list<ModuleLoader*> ModuleLoader::Audio_ADCs;

ModuleLoader::ModuleLoader(const char* name, int moduleId,int argc,char **argv): ID(moduleId) {
    OSErr error;
    isStartPoint = false;
    isADC = false;
    timingRef = 0L;
    char *modName2 = (char*)calloc(strlen(name)+10,sizeof(char));
    strcpy(modName2,name);
    strcat(modName2,".panda");
    CFStringRef stringa = CFStringCreateWithCString(kCFAllocatorSystemDefault,modName2,NULL);
    error = CreateBundleFromName(stringa, &module);
    if ((error == noErr) && (module != NULL)) { 
        status = 1; 
        loadBundle(ID,argc,argv); 
        Connection::addModule(ID,this); 
        if(strcmp(name,"Audio/StartPoint")==0 || strcmp(name,"audio/startpoint")==0) { ModuleLoader::Audio_StartPoints.push_front(this); isStartPoint = true; }
        if(strcmp(name,"Audio/ADC")==0 || strcmp(name,"audio/adc")==0) { ModuleLoader::Audio_ADCs.push_front(this); isADC = true; }
        modName = (char*)calloc(strlen(name)+2,sizeof(char));
        memset(modName,0x0,sizeof(char)*strlen(name)+2);
        strcpy(modName,name);
        free(modName2);
        theArgc = argc;
        if(argc>0) {
            theArgv = (char**)malloc(sizeof(char*)*argc);
            for(int i = 0;i<argc;i++) {
                theArgv[i] = (char*)calloc(strlen(argv[i]),sizeof(char));
                strcpy(theArgv[i],argv[i]);
            }
        } else theArgv = NULL;
    } else free(modName2);
}

ModuleLoader::~ModuleLoader() {
    releaseBundle();
}

OSErr ModuleLoader::CreateBundleFromName(CFStringRef theFrameworkName, CFBundleRef *theBundle)
{
	OSErr theErr;
	FSRef theRef;
	CFURLRef theFrameworkURL;
	CFURLRef theBundleURL;
        
	
        theErr = FSPathMakeRef((UInt8*)"/Library/Application Support/Panda/Modules/",&theRef,NULL);
	if (theErr == noErr)
	{
		
		theFrameworkURL = CFURLCreateFromFSRef(kCFAllocatorSystemDefault, &theRef);
		
		if (theFrameworkURL != NULL)
		{
			theBundleURL = CFURLCreateCopyAppendingPathComponent(kCFAllocatorSystemDefault,
				theFrameworkURL, theFrameworkName, false);
			
			CFRelease(theFrameworkURL);
			
			if (theBundleURL != NULL)
			{
                            *theBundle = CFBundleCreate(kCFAllocatorSystemDefault, theBundleURL);
                            CFRelease(theBundleURL);
			} else theErr = -55;
		}
	}
        
        if(theErr==-55 || module==NULL) {
            FSRef homeFolderRef;
            CFURLRef homeUrl;
            
            theErr = FSFindFolder(kUserDomain,kCurrentUserFolderType,kDontCreateFolder,&homeFolderRef);
            
            if(theErr==noErr) {
                
                homeUrl = CFURLCreateFromFSRef(kCFAllocatorSystemDefault,&homeFolderRef);
                
                if(homeUrl) {
                    
                    char *urlStr = (char*)calloc(256,sizeof(char));
                    CFURLGetFileSystemRepresentation(homeUrl,FALSE,(UInt8*)urlStr,256);
                    strcat(urlStr,"/Library/Application Support/Panda/Modules/");
                    
                    theErr = FSPathMakeRef((UInt8*)urlStr,&theRef,NULL);
                    
                    free(urlStr);
                    
                    if (theErr == noErr)
                    {
                        theFrameworkURL = CFURLCreateFromFSRef(kCFAllocatorSystemDefault, &theRef);
		
                        if (theFrameworkURL != NULL)
                        {
                                theBundleURL = CFURLCreateCopyAppendingPathComponent(kCFAllocatorSystemDefault,
                                    theFrameworkURL, theFrameworkName, false);
			
                                CFRelease(theFrameworkURL);
			
                                if (theBundleURL != NULL)
                                {
                                    *theBundle = CFBundleCreate(kCFAllocatorSystemDefault, theBundleURL);
                                    CFRelease(theBundleURL);
                                    if(theBundle!=NULL) theErr = noErr;
                                    else theErr = -55;
                                } else theErr = -55;
                        }
                    }
                }
            }	
        }

        
        if(theErr==-55 || module==NULL) {
            
            CFBundleRef auBund = CFBundleGetMainBundle();
            CFURLRef resUrl = CFBundleCopyResourcesDirectoryURL(auBund);
            
            theErr = noErr;
            
            if(theErr==noErr) {
                
                if(resUrl) {
                    
                    char *urlStr = (char*)calloc(256,sizeof(char));
                    CFURLGetFileSystemRepresentation(resUrl,FALSE,(UInt8*)urlStr,256);
                    strcat(urlStr,"/Modules/");
                    
                    theErr = FSPathMakeRef((UInt8*)urlStr,&theRef,NULL);
                    
                    free(urlStr);
                    
                    if (theErr == noErr)
                    {
                        theFrameworkURL = CFURLCreateFromFSRef(kCFAllocatorSystemDefault, &theRef);
		
                        if (theFrameworkURL != NULL)
                        {
                                theBundleURL = CFURLCreateCopyAppendingPathComponent(kCFAllocatorSystemDefault,
                                    theFrameworkURL, theFrameworkName, false);
			
                                CFRelease(theFrameworkURL);
			
                                if (theBundleURL != NULL)
                                {
                                    *theBundle = CFBundleCreate(kCFAllocatorSystemDefault, theBundleURL);
                                    CFRelease(theBundleURL);
                                    if(theBundle!=NULL) theErr = noErr;
                                    else theErr = -55;
                                } else theErr = -55;
                        }
                    }
                }
            }
            CFRelease(resUrl);
        }
	
	return theErr;
}

void ModuleLoader::loadBundle(int ID,int argc,char **argv) {
	
        OSErr isLoaded;
	
	if ((status == 1) && (module != NULL))
	{
                if(CFBundleIsExecutableLoaded(module)) { 
                #ifdef DEBUG
                    printf("bundle already loaded\n"); 
                #endif
                    isLoaded=true; 
                }
		else isLoaded = CFBundleLoadExecutable(module);		
		
		if (isLoaded)
		{	
			functs.NewPandaMod = (void *) CFBundleGetFunctionPointerForName(module, CFSTR("HelloPanda"));
                        functs.InletCallBackPtr = (void *) CFBundleGetFunctionPointerForName(module, CFSTR("InletCallback"));
                        functs.GetParameterListPtr = (void *) CFBundleGetFunctionPointerForName(module, CFSTR("GetMethodsList"));
                        functs.flushModulePtr = (void *) CFBundleGetFunctionPointerForName(module, CFSTR("FlushModule"));
                        int err;
                        data = functs.NewPandaMod(&err,ID,argc,argv);
                        TYPE moduleType = functs.GetParameterListPtr(data).moduleType;
                        outlets = functs.GetParameterListPtr(data).outlets;
                        inlets = functs.GetParameterListPtr(data).inlets;
                        switch(moduleType) {
                            case AUDIO_PARAMETERS:
                            #ifdef DEBUG
                                printf("AUDIO_PARAMETERS\n");
                            #endif
                                functs.AudioProcessPtr = (void *) CFBundleGetFunctionPointerForName(module, CFSTR("AudioProcess"));
                                audio_outlets = functs.GetParameterListPtr(data).audio_outlets;
                                audio_inlets = functs.GetParameterListPtr(data).audio_inlets;
                                if(audio_inlets>0) {                                    
                                    functs.AllocInletsPtr = (void *) CFBundleGetFunctionPointerForName(module, CFSTR("AllocInlets"));
                                    inletsBufferList.buffers = functs.AllocInletsPtr(data,AudioRender::gBufferSize,&inletsBufferList.manyBuffers);
                                } else printf("cannot call AllocInlets\n");
                                break;
                            case ONLY_PARAMETERS:
                            #ifdef DEBUG
                                printf("ONLY_PARAMETERS\n");
                            #endif
                                break;
                        }
                        if(err!=0) { status = -1; }
                }
	}
}

void ModuleLoader::releaseBundle() {
    if(status==1 && module!=NULL && data!=NULL) {
    
        if(isADC) {
            list<ModuleLoader*>::iterator i;
            for(i=ModuleLoader::Audio_ADCs.begin(); i!=ModuleLoader::Audio_ADCs.end();++i) {
                ModuleLoader *theConn = *i;
                if(theConn->ID == ID) { ModuleLoader::Audio_ADCs.erase(i); break; }
            }
        }
        else if(isStartPoint) {
            list<ModuleLoader*>::iterator i;
            for(i=ModuleLoader::Audio_StartPoints.begin(); i!=ModuleLoader::Audio_StartPoints.end();++i) {
                ModuleLoader *theConn = *i;
                if(theConn->ID == ID) { ModuleLoader::Audio_StartPoints.erase(i); break; }
            }
        }
    
        Connection::removeModule(ID);
    
        map<int,ModuleLoader*>::iterator ModulePos;
        for(ModulePos = Connection::modules.begin();ModulePos != Connection::modules.end(); ++ModulePos) {
            if(ModulePos->second!=NULL) {
                ModuleLoader *theModule = ModulePos->second;
                if(strcmp(modName,theModule->modName)==0) {
                    functs.flushModulePtr(data);
                    free(modName);
                    data = NULL;
                    return;
                }
            }
        }
        
        functs.flushModulePtr(data);
        
        if(functs.GetParameterListPtr(data).manyMethods!='coco') {
   
        if(CFBundleIsExecutableLoaded(module)) CFBundleUnloadExecutable(module);
    
        CFRelease(module);
        
        } else printf("COCOA MODULE not released\n");
    
        free(modName);
        
        if(theArgc>0 && theArgv) {
            for(int i = 0;i<theArgc;i++) {
                free(theArgv[i]);
            }
            free(theArgv);
        } 

        data = NULL;
        
        
    #ifdef DEBUG
        printf("Bundle released\n");
    #endif
    }
}

void ModuleLoader::executeMethod(int inlet, Var *inValues) {
    functs.InletCallBackPtr(data,inlet,inValues);
}

int ModuleLoader::audio_executeMethod(long bufferSize) {
    int res = 0;
    
    res = functs.AudioProcessPtr(data,bufferSize);
    if(res==0)Connection::audio_executeConnected(ID,0,bufferSize);
    
    /*
    if(timing!=timingRef) {
        res = functs.AudioProcessPtr(data,bufferSize);
        if(res==0) {
            for(int i=0;i<audio_outlets;i++) {
                Connection::audio_executeConnected(ID,i,bufferSize);
            }   
        }
        timingRef = timing;
    }
    */
    return res;
}

/*
void ModuleLoader::printMethods() {
    ParamList list;
    list = functs.GetParameterListPtr(data);
    int many = list.manyMethods;
    printf("-----------------------\n");
    printf("Module \"%s\" methods:\n",modName);
    for(int i=0;i<many;i++) {
        printf("%d): \"%s\" ",i,list.functions[i]);
        switch(*list.types[i]) {
            case INT_T:
                printf("of kind: \"INT\" ");
                break;
            case CHAR_T:
                printf("of kind: \"CHAR\" ");
                break;
            case FLOAT_T:
                printf("of kind: \"FLOAT\" ");
                break;
            case DOUBLE_T:
                printf("of kind: \"DOUBLE\" ");
                break;
            case BYTE_T:
                printf("of kind: \"BYTE\" ");
                break;
            case ULONG_T:
                printf("of kind: \"UNSIGNED LONG\" ");
                break;
            case LONG_T:
                printf("of kind: \"LONG\" ");
                break;
            case LDOUBLE_T:
                printf("of kind: \"LONG DOUBLE\" ");
                break;
            case OBJECT_T:
                printf("of kind: \"OBJECT\" ");
                break;
            case PTR_T:
                printf("of kind: \"POINTER\" ");
                break;
            case BANG_T:
                printf("of kind: \"BANG\" ");
                break;
        }
        printf("\n");
    }
    printf("-----------------------\n");
}

char * ModuleLoader::getMethodName(int n) {
    ParamList list;
    list = functs.GetParameterListPtr(data);
    if(list.functions[n]!=NULL) { 
        return list.functions[n];
    }
    return NULL;
}

int ModuleLoader::getMethodKind(int n) {
    ParamList list;
    list = functs.GetParameterListPtr(data);
    if(list.types[n]!=NULL) { 
        return *list.types[n];
    }
    return -1;
}

int ModuleLoader::getMethodNumber(char *name) {
    ParamList list;
    list = functs.GetParameterListPtr(data);
    for(int i = 0;i<list.manyMethods;i++) {
        if(strcmp(list.functions[i],name) == 0) return i;
    }
    return -1;
}
*/

CFBundleRef ModuleLoader::GetModuleBundle(int moduleID) {
    ModuleLoader *Module = Connection::moduleFromID(moduleID);
    return Module->module;
}

//outlet callback
/*#ifdef __cplusplus
extern "C" {
#endif*/

CFBundleRef GetModuleBundleFromID(int ID) {
    return ModuleLoader::GetModuleBundle(ID);
}

list<S_SharedPtr> keysList;

int setSharedPointer(char *key,void *ptr) {
    S_SharedPtr newItem;
    strcpy(&newItem.key[0],key);
    newItem.ptr = ptr;
    
    keysList.push_front(newItem);
    
    return TRUE;
}

void *sharedPointer(char *key) {
    list<S_SharedPtr>::iterator i;
    for(i=keysList.begin(); i!=keysList.end();++i) {
        S_SharedPtr item = *i;
        if(strcmp(&item.key[0],key)==0) {
            return item.ptr;
        }
    }

    return NULL;
}

void removeSharedPointer(char *key) {
    list<S_SharedPtr>::iterator i;
    for(i=keysList.begin(); i!=keysList.end();++i) {
        S_SharedPtr item = *i;
        if(strcmp(&item.key[0],key)==0) {
            keysList.erase(i);
            return;
        }
    }
}

void PandaLog(char *fmt,...) {
    va_list ap;
    va_start(ap, fmt);
    fprintf(stderr,"PandaLog: ");
    vfprintf(stderr, fmt, ap);
    va_end(ap);
}

void outletCall(int ID, int outletNumber, void * values) {
    Var *variables = (Var*)values;
    Connection::executeConnected(ID,outletNumber,variables);
}

int audio_StartPointCall(int ID, int outletNumber,long bufferSize) {
    return Connection::audio_executeConnected(ID,outletNumber,bufferSize);
}

InletsBuffers2 GetOutletBuffers(int ID, int outletNumber) {
    InletsBuffers2 buffersList;
    Connection::getBuffersConnected(ID,outletNumber,&buffersList);
    return buffersList;
}

float **GetADCBuffers(int *channels) {
    *channels = AudioRender::gInputChannels;
    return AudioRender::getADC();
}

float **GetDACBuffers(int *channels) {
    *channels = AudioRender::gOutputChannels;
    return AudioRender::getDAC();
}

long GetBufferSize() {
    long res = AudioRender::gBufferSize;
    return res;
}

float GetSampleRate() {
    float res = AudioRender::gSampleRate;
    return res;
}
int GetDeviceInputChannels() {
    int res = AudioRender::gInputChannels;
    return res;
}
int GetDeviceOutputChannels() {
    int res = AudioRender::gOutputChannels;
    return res;
}

const AudioTimeStamp * GetDeviceTime() {
    const AudioTimeStamp *res = AudioRender::gTime;
    return res;
}

float **GetPluginADCBuffers(char *istanceName) {
    return AudioUnitRender::getADC(istanceName);
}

float **GetPluginDACBuffers(char *istanceName) {
    return AudioUnitRender::getDAC(istanceName);
}

P_BOOL GetPluginAudioConfiguration(char *istanceName,float *outSR,long *outBSize,int *outInCh,int *outOutCh) {
    return AudioUnitRender::getAudioConfiguration(istanceName,outSR,outBSize,outInCh,outOutCh);
}

int c_bool(P_BOOL the_p_bool) {
    if(the_p_bool==B_YES) return TRUE;
    return FALSE;
}

P_BOOL p_bool(int c_cond) {
    if(c_cond) return B_YES;
    return B_NO;
}

/*#ifdef __cplusplus
}
#endif*/


//class connection
map<int,ModuleLoader*>Connection::modules;
map<int,ModuleLoader*>::iterator Connection::pos;
list<Connection*> Connection::connections;
list<Connection*> Connection::audio_connections;


Connection::Connection(int fromID, int toID, int fromOutlet, int toInlet) : IDfrom(fromID),IDto(toID),outletFrom(fromOutlet),inletTo(toInlet) {
    
}

bool Connection::connect(int fromID,int fromOutlet,int toID,int toInlet) {
    list<Connection*>::iterator i;
    for(i=Connection::connections.begin(); i!=Connection::connections.end();++i) {
        Connection *theConn = *i;
        if(theConn->IDfrom==fromID && theConn->IDto==toID && theConn->outletFrom==fromOutlet && theConn->inletTo==toInlet) return false;
    }
    Connection *conn = new Connection(fromID,toID,fromOutlet,toInlet);
    Connection::connections.push_front(conn);
    return true;
}

bool Connection::savePandaStatusToFile(const char *fileName,int code1,int code2,int code3,int code4) {
    if(code1!=CODE_1 && code2!=CODE_2 && code3!=CODE_3 && code4!=CODE_4) return false;
    
    int toFile = open(fileName,BINCREATE,0666);
    if(toFile==-1) return false;
    int id;
    
    int audioStat = 6667;
    write(toFile,(void*)&audioStat,sizeof(int));
    
    id = 'modu';
    write(toFile,(void*)&id,sizeof(int));
    int manyMod = Connection::modules.size();
    write(toFile,(void*)&manyMod,sizeof(int));
    
    map<int,ModuleLoader*>::iterator it;
    
    for(it=Connection::modules.begin(); it!=Connection::modules.end();++it) {
        if(it->second) {
            ModuleLoader *theConn = it->second;
            int strLen = strlen(theConn->modName);
            write(toFile,(void *)&strLen,sizeof(int));
            write(toFile,(void *)theConn->modName,sizeof(char)*strLen);
            
            char audioCheck[6];
            for(int c = 0;c<6;c++) {
                if(c==5) { audioCheck[c] = '\0'; break; }
                audioCheck[c] = theConn->modName[c];
            }
            if(strcmp(&audioCheck[0],"audio")==0 || strcmp(&audioCheck[0],"Audio")==0) audioStat = 6666;
            
            int ID = theConn->getID();
            write(toFile,(void *)&ID,sizeof(int));
            write(toFile,(void *)&theConn->theArgc,sizeof(int));
            for(int beta=0;beta<theConn->theArgc;beta++) {
                int strLen2 = strlen(theConn->theArgv[beta]);
                write(toFile,(void*)&strLen2,sizeof(int));
                write(toFile,(void*)theConn->theArgv[beta],sizeof(char)*strLen2);
            }
        }
    }
    
    list<Connection*>::iterator i;
    //non audio connections
    
    id = 'norm';
    write(toFile,(void*)&id,sizeof(int));
    int manyConn = Connection::connections.size();
    write(toFile,(void*)&manyConn,sizeof(int));
    for(i=Connection::connections.begin(); i!=Connection::connections.end();++i) {
        Connection *theConn = *i;
        write(toFile,(void *)theConn,sizeof(Connection));
    }
    PandaLog("I've saved %d connections\n",manyConn);
    
    bool deleteAfter = false;
    if(manyConn==0) deleteAfter = true;
    
    //audio connections
    id = 'audi';
    write(toFile,(void*)&id,sizeof(int));
    manyConn = Connection::audio_connections.size();
    write(toFile,(void*)&manyConn,sizeof(int));
    for(i=Connection::audio_connections.begin(); i!=Connection::audio_connections.end();++i) {
        Connection *theConn = *i;
        write(toFile,(void *)theConn,sizeof(Connection));
    }
    PandaLog("I've saved %d audio connections\n",manyConn);
    
    if(manyConn>0) deleteAfter = false;
    
    pwrite(toFile,(void*)&audioStat,sizeof(int),0);
    
    if(deleteAfter) {
        //??
    }
    
    close(toFile);
    
    return true;
}

bool Connection::loadPandaStatusFromFile(const char *fileName,int code1,int code2,int code3,int code4,loadStatusAudioCallbackPtr call) {
    if(code1!=CODE_1 && code2!=CODE_2 && code3!=CODE_3 && code4!=CODE_4) return false;
    
    int toFile = open(fileName,O_RDONLY,0666);
    if(toFile==-1) return false;
    
    int id = 0;
    if(read(toFile,(void*)&id,sizeof(int))!=sizeof(int)) return false;
    if(read(toFile,(void*)&id,sizeof(int))!=sizeof(int)) return false;
    if(id!='modu') return false;
    int manyMod = 0;
    if(read(toFile,(void*)&manyMod,sizeof(int))!=sizeof(int)) return false;
    
    if(manyMod<=0) return false;
    
    PandaLog("I'm loading %d modules\n",manyMod);
    
    for(int i=0;i<manyMod;i++) {
        
        int strLen = 0;
        if(read(toFile,(void *)&strLen,sizeof(int))!=sizeof(int)) return false;
        char *name = (char*)calloc(strLen,sizeof(char));
        int size = sizeof(char)*strLen;
        if(read(toFile,(void *)name,sizeof(char)*strLen)!=size) return false;
        int modID;
        if(read(toFile,(void *)&modID,sizeof(int))!=sizeof(int)) return false;
        //PandaLog("module: %s id: %d name length: %d\n",name,modID,strLen);
    
        int argc;
        if(read(toFile,(void *)&argc,sizeof(int))!=sizeof(int)) return false;
        char **argv = NULL;
        if(argc>0) argv = (char**)malloc(sizeof(char*)*argc);
        else argv = NULL;
        for(int beta=0;beta<argc;beta++) {
            int strLen2 = 0;
            if(read(toFile,(void*)&strLen2,sizeof(int))!=sizeof(int)) return false;
            argv[beta] = (char*)calloc(strLen2,sizeof(char));
            if(read(toFile,(void*)argv[beta],sizeof(char)*strLen2) != sizeof(char)*strLen2) return false;
        }

        char *modName = (char*)calloc(strLen+1,sizeof(char));
        memcpy(modName,name,sizeof(char)*strLen);
        
        modName[strLen] = '\0';
        
        PandaLog("module: %s length: %d\n",modName,strLen);
        
        
        ModuleLoader *module = new ModuleLoader(modName,modID,argc,argv);
        free(name); free(modName);
        
        if(argv) {
            for(int b=0;b<argc;b++) {
                free(argv[b]);
            }
            free(argv);
        }
    }
    
    if(read(toFile,(void*)&id,sizeof(int))!=sizeof(int)) return false;
    if(id!='norm') return false;
    
    int manyConn = 0;
    if(read(toFile,(void*)&manyConn,sizeof(int))!=sizeof(int)) return false;
    
    PandaLog("There are %d connections to restore\n",manyConn);
    
    for(int i=0;i<manyConn;i++) {
        Connection *conn = (Connection*)malloc(sizeof(Connection));
        if(read(toFile,(void *)conn,sizeof(Connection))!=sizeof(Connection)) return false;
        if(!conn) return false;
        PandaLog("connection %d : from %d outlet %d to %d inlet %d\n",i,conn->IDfrom,conn->outletFrom,conn->IDto,conn->inletTo);
        Connection *c = new Connection(conn->IDfrom,conn->IDto,conn->outletFrom,conn->inletTo);
        Connection::connections.push_front(c);
        free(conn);
    }
    
    if(read(toFile,(void*)&id,sizeof(int))!=sizeof(int)) return false;
    if(id!='audi') return false;
    
    manyConn = 0;
    if(read(toFile,(void*)&manyConn,sizeof(int))!=sizeof(int)) return false;
    
    PandaLog("There are %d audio connections to restore\n",manyConn);
    
    for(int i=0;i<manyConn;i++) {
        Connection *conn = (Connection*)malloc(sizeof(Connection));
        if(read(toFile,(void *)conn,sizeof(Connection))!=sizeof(Connection)) return false;
        if(!conn) return false;
        PandaLog("connection %d : from %d outlet %d to %d inlet %d\n",i,conn->IDfrom,conn->outletFrom,conn->IDto,conn->inletTo);
        Connection *c = new Connection(conn->IDfrom,conn->IDto,conn->outletFrom,conn->inletTo);
        Connection::audio_connections.push_front(c);
        free(conn);
    }
    
    close(toFile);
    PandaLog("Panda status loaded\n");
    return true;
}

bool Connection::savePandaStatusToFileForPlugin(const char *fileName,int code1,int code2,int code3,int code4) {
    if(code1!=CODE_1 && code2!=CODE_2 && code3!=CODE_3 && code4!=CODE_4) return false;
    
    int toFile = open(fileName,BINCREATE,0666);
    if(toFile==-1) return false;
    int id;
    

    
    int audioStat = 6667;
    write(toFile,(void*)&audioStat,sizeof(int));
    
    int guiStat = 5555;
    write(toFile,(void*)&guiStat,sizeof(int));
    
    int parameters = 0;
    int *nPar = (int*)sharedPointer("ElCa-Plugin-ParametersN");
    if(nPar) parameters = *nPar;
    write(toFile,(void*)&parameters,sizeof(int));
    
    id = 'modu';
    write(toFile,(void*)&id,sizeof(int));
    int manyMod = Connection::modules.size();
    write(toFile,(void*)&manyMod,sizeof(int));
    
    map<int,ModuleLoader*>::iterator it;
    
    for(it=Connection::modules.begin(); it!=Connection::modules.end();++it) {
        if(it->second) {
            ModuleLoader *theConn = it->second;
            int strLen = strlen(theConn->modName);
            write(toFile,(void *)&strLen,sizeof(int));
            write(toFile,(void *)theConn->modName,sizeof(char)*strLen);
            
            char audioCheck[6];
            for(int c = 0;c<6;c++) {
                if(c==5) { audioCheck[c] = '\0'; break; }
                audioCheck[c] = theConn->modName[c];
            }
            if(strcmp(&audioCheck[0],"audio")==0 || strcmp(&audioCheck[0],"Audio")==0) audioStat = 6666;
            
            char guiCheck[4];
            for(int c = 0;c<4;c++) {
                if(c==3) { guiCheck[c] = '\0'; break; }
                guiCheck[c] = theConn->modName[c];
            }
            if(strcmp(&guiCheck[0],"GUI")==0 || strcmp(&guiCheck[0],"gui")==0) guiStat = 5556;
            
            int ID = theConn->getID();
            write(toFile,(void *)&ID,sizeof(int));
            write(toFile,(void *)&theConn->theArgc,sizeof(int));
            for(int beta=0;beta<theConn->theArgc;beta++) {
                int strLen2 = strlen(theConn->theArgv[beta]);
                write(toFile,(void*)&strLen2,sizeof(int));
                write(toFile,(void*)theConn->theArgv[beta],sizeof(char)*strLen2);
            }
        }
    }
    
    list<Connection*>::iterator i;
    //non audio connections
    
    id = 'norm';
    write(toFile,(void*)&id,sizeof(int));
    int manyConn = Connection::connections.size();
    write(toFile,(void*)&manyConn,sizeof(int));
    for(i=Connection::connections.begin(); i!=Connection::connections.end();++i) {
        Connection *theConn = *i;
        write(toFile,(void *)theConn,sizeof(Connection));
    }
    PandaLog("I've saved %d connections\n",manyConn);
    
    bool deleteAfter = false;
    if(manyConn==0) deleteAfter = true;
    
    //audio connections
    id = 'audi';
    write(toFile,(void*)&id,sizeof(int));
    manyConn = Connection::audio_connections.size();
    write(toFile,(void*)&manyConn,sizeof(int));
    for(i=Connection::audio_connections.begin(); i!=Connection::audio_connections.end();++i) {
        Connection *theConn = *i;
        write(toFile,(void *)theConn,sizeof(Connection));
    }
    PandaLog("I've saved %d audio connections\n",manyConn);
    
    if(manyConn>0) deleteAfter = false;
    
    pwrite(toFile,(void*)&audioStat,sizeof(int),0);
    pwrite(toFile,(void*)&guiStat,sizeof(int),1*sizeof(int));
    pwrite(toFile,(void*)&parameters,sizeof(int),2*sizeof(int));
    
    if(deleteAfter) {
        //??
    }
    
    close(toFile);
    
    return true;

}

bool Connection::loadPandaStatusFromFileForPlugin(const char *fileName,const char *instanceName,int code1,int code2,int code3,int code4,loadStatusAudioCallbackPtr call) {
    if(code1!=CODE_1 && code2!=CODE_2 && code3!=CODE_3 && code4!=CODE_4) return false;
    
    int K = AudioUnitRender::pluginsModulesK;
    
    int toFile = open(fileName,O_RDONLY,0666);
    if(toFile==-1) return false;
    
    int id = 0;
    if(read(toFile,(void*)&id,sizeof(int))!=sizeof(int)) return false;
    if(read(toFile,(void*)&id,sizeof(int))!=sizeof(int)) return false;
    if(read(toFile,(void*)&id,sizeof(int))!=sizeof(int)) return false;
    if(read(toFile,(void*)&id,sizeof(int))!=sizeof(int)) return false;
    if(id!='modu') return false;
    int manyMod = 0;
    if(read(toFile,(void*)&manyMod,sizeof(int))!=sizeof(int)) return false;
    
    if(manyMod<=0) return false;
    
    PandaLog("I'm loading %d modules\n",manyMod);
    
    for(int i=0;i<manyMod;i++) {
        
        int strLen = 0;
        if(read(toFile,(void *)&strLen,sizeof(int))!=sizeof(int)) return false;
        char *name = (char*)calloc(strLen,sizeof(char));
        int size = sizeof(char)*strLen;
        if(read(toFile,(void *)name,size)!=size) return false;
        int modID;
        if(read(toFile,(void *)&modID,sizeof(int))!=sizeof(int)) return false;
        
        modID+=K;
        
        int argc;
        if(read(toFile,(void *)&argc,sizeof(int))!=sizeof(int)) return false;
        char **argv = NULL;
        if(argc>0) argv = (char**)malloc(sizeof(char*)*argc);
        else argv = NULL;
        for(int beta=0;beta<argc;beta++) {
            int strLen2 = 0;
            if(read(toFile,(void*)&strLen2,sizeof(int))!=sizeof(int)) return false;
            argv[beta] = (char*)calloc(strLen2,sizeof(char));
            if(read(toFile,(void*)argv[beta],sizeof(char)*strLen2) != sizeof(char)*strLen2) return false;
        }

        char *modName = (char*)calloc(strLen,sizeof(char));
        memcpy(modName,name,sizeof(char)*strLen);
        
        modName[strLen] = '\0';
        
        PandaLog("module: %s id: %d name length: %d\n",modName,modID,strLen);
        
        if(strcmp(modName,"Audio/DAC")!=0 && strcmp(modName,"audio/dac")!=0 && strcmp(modName,"Audio/ADC")!=0 && strcmp(modName,"audio/adc")!=0) {
        
        
        ModuleLoader *module = new ModuleLoader(modName,modID,argc,argv);
        if(module->IsLoaded()) 
        {
        
        AudioUnitRender::addModuleAtInstance(modID,(char*)instanceName);
        
        if(strcmp(modName,"Audio/StartPoint")==0 || strcmp(modName,"audio/startpoint")==0 ) AudioUnitRender::addStartPoint((char*)instanceName,module);
        else if(strcmp(modName,"Audio/PL-ADC")==0 || strcmp(modName,"audio/pl-adc")==0 ) AudioUnitRender::addADC((char*)instanceName,module);
        
        }
        
        }
        
        if(modID>AudioUnitRender::pluginsModulesK) AudioUnitRender::pluginsModulesK = modID;
        
        free(name); free(modName);
        
        if(argv) {
            for(int b=0;b<argc;b++) {
                free(argv[b]);
            }
            free(argv);
        }
    }
    
    if(read(toFile,(void*)&id,sizeof(int))!=sizeof(int)) return false;
    if(id!='norm') return false;
    
    int manyConn = 0;
    if(read(toFile,(void*)&manyConn,sizeof(int))!=sizeof(int)) return false;
    
    PandaLog("There are %d connections to restore\n",manyConn);
    
    for(int i=0;i<manyConn;i++) {
        Connection *conn = (Connection*)malloc(sizeof(Connection));
        if(read(toFile,(void *)conn,sizeof(Connection))!=sizeof(Connection)) return false;
        if(!conn) return false;
        conn->IDfrom += K;
        conn->IDto += K;
        PandaLog("connection %d : from %d outlet %d to %d inlet %d\n",i,conn->IDfrom,conn->outletFrom,conn->IDto,conn->inletTo);
        Connection *c = new Connection(conn->IDfrom,conn->IDto,conn->outletFrom,conn->inletTo);
        Connection::connections.push_front(c);
        free(conn);
    }
    
    if(read(toFile,(void*)&id,sizeof(int))!=sizeof(int)) return false;
    if(id!='audi') return false;
    
    manyConn = 0;
    if(read(toFile,(void*)&manyConn,sizeof(int))!=sizeof(int)) return false;
    
    PandaLog("There are %d audio connections to restore\n",manyConn);
    
    for(int i=0;i<manyConn;i++) {
        Connection *conn = (Connection*)malloc(sizeof(Connection));
        if(read(toFile,(void *)conn,sizeof(Connection))!=sizeof(Connection)) return false;
        if(!conn) return false;
        conn->IDfrom += K;
        conn->IDto += K;
        PandaLog("connection %d : from %d outlet %d to %d inlet %d\n",i,conn->IDfrom,conn->outletFrom,conn->IDto,conn->inletTo);
        Connection *c = new Connection(conn->IDfrom,conn->IDto,conn->outletFrom,conn->inletTo);
        Connection::audio_connections.push_front(c);
        free(conn);
    }
    
    close(toFile);
    PandaLog("Panda status loaded\n");
    return true;

}

bool Connection::disconnect(int fromID,int fromOutlet,int toID,int toInlet) {
    list<Connection*>::iterator i;
    for(i=Connection::connections.begin(); i!=Connection::connections.end();++i) {
        Connection *theConn = *i;
        if(theConn->IDfrom==fromID && theConn->IDto==toID && theConn->outletFrom==fromOutlet && theConn->inletTo==toInlet) { 
            Connection::connections.erase(i);
            delete theConn;
            return true;
        }
    }
    return false;
}

bool Connection::audio_connect(int fromID,int fromOutlet,int toID,int toInlet) {
    list<Connection*>::iterator i;
    for(i=Connection::audio_connections.begin(); i!=Connection::audio_connections.end();++i) {
        Connection *theConn = *i;
        if(theConn->IDfrom==fromID && theConn->IDto==toID && theConn->outletFrom==fromOutlet && theConn->inletTo==toInlet) return false;
    }
    Connection *conn = new Connection(fromID,toID,fromOutlet,toInlet);
    Connection::audio_connections.push_front(conn);
    return true;
}

bool Connection::audio_disconnect(int fromID,int fromOutlet,int toID,int toInlet) {
    list<Connection*>::iterator i;
    for(i=Connection::audio_connections.begin(); i!=Connection::audio_connections.end();++i) {
        Connection *theConn = *i;
        if(theConn->IDfrom==fromID && theConn->IDto==toID && theConn->outletFrom==fromOutlet && theConn->inletTo==toInlet) { 
            Connection::audio_connections.erase(i);
            delete theConn;
            return true;
        }
    }
    return false;
}

int Connection::executeConnected(int fromID,int fromOutlet,Var *values) {
    list<Connection*>::iterator i;
    for(i=Connection::connections.begin(); i!=Connection::connections.end();++i) {
        Connection *theConn = *i;
        if(theConn->IDfrom==fromID && theConn->outletFrom==fromOutlet) {
            pos = Connection::modules.find(theConn->IDto);
            if(pos->second==NULL) break;
            ModuleLoader *theModule = pos->second;
            theModule->executeMethod(theConn->inletTo,values);
        }
    }
    
    return 0;
}

void Connection::getBuffersConnected(int fromID,int fromOutlet,InletsBuffers2 *outList) {
    list<Connection*>::iterator i;
    int count = 0;
    outList->manyBuffers = 0;
    for(i=Connection::audio_connections.begin(); i!=Connection::audio_connections.end();++i) {
        Connection *theConn = *i;
        if(theConn->IDfrom==fromID && theConn->outletFrom==fromOutlet) {
            pos = Connection::modules.find(theConn->IDto);
            if(pos->second==NULL) break;
            ModuleLoader *theModule = pos->second;
            outList->buffers[count] = theModule->inletsBufferList.buffers[theConn->inletTo];
            count++;
            outList->manyBuffers++;
        }
    }
}

bool isDone(int toID,int *array, int many) {
    int i;
    for(i=0;i<many;i++) {
        if(array[i]==toID) return true;
    }
    return false;
}


int Connection::audio_executeConnected(int fromID,int fromOutlet,long bufferSize) {
    list<Connection*>::iterator i;
    
    int done[256];
    int doneCount = 0;
    
    for(i=Connection::audio_connections.begin(); i!=Connection::audio_connections.end();++i) {
        Connection *theConn = *i;
        if(theConn->IDfrom==fromID /*&& theConn->outletFrom==fromOutlet*/) {
            pos = Connection::modules.find(theConn->IDto);
            if(pos->second==NULL) break;
            if(!isDone(theConn->IDto,&done[0],doneCount)) {
                done[doneCount] = theConn->IDto;
                doneCount++;
                ModuleLoader *theModule = pos->second;
                theModule->audio_executeMethod(bufferSize);
            } 
        }
    }
        
    return -1;
}

void Connection::addModule(int ID,ModuleLoader *module) {
    Connection::modules[ID] = module;
}

void Connection::removeModule(int ID) {
    Connection::modules.erase(ID);
}

ModuleLoader * Connection::moduleFromID(int ID) {
    Connection::pos = Connection::modules.find(ID);
    return Connection::pos->second;
}


//Audio render CLASS

float AudioRender::gSampleRate;
long AudioRender::gBufferSize;
int AudioRender::gInputChannels;
int AudioRender::gOutputChannels;
AudioRender *AudioRender::theRender = NULL;
bool AudioRender::isProcessing = false;
const AudioTimeStamp *AudioRender::gTime;

AudioRender::AudioRender(float sampleRate,long bufferSize,int inChannels, int outChannels, char *device) : 	vSampleRate(sampleRate),vBufferSize(bufferSize) 
{
    inBuffers=NULL;
    outBuffers=NULL;
    status = ConfigureAudioProc(sampleRate,bufferSize,outChannels,inChannels,device);
    isProcessing = false;
    if(status) {
        inBuffers = (float**)malloc(sizeof(float*)*vChannels);
        outBuffers = (float**)malloc(sizeof(float*)*vChannels);
        for(int i =0;i< vChannels;i++) {
            inBuffers[i] = (float*)malloc(sizeof(float)*vBufferSize);
        }
        for(int i =0;i< vChannels;i++) {
            outBuffers[i] = (float*)malloc(sizeof(float)*vBufferSize);
            memset(outBuffers[i], 0, vBufferSize*sizeof(float));
        }
    }
    
    AudioRender::gSampleRate = vSampleRate;
    AudioRender::gBufferSize = vBufferSize;
    AudioRender::gInputChannels = vInChannels;
    AudioRender::gOutputChannels = vChannels;
    AudioRender::theRender = this;
}

AudioRender::~AudioRender() {
if(status) {
    if(isProcessing) AudioDeviceStop(vDevice,process);
    OSStatus err = AudioDeviceRemoveIOProc(vDevice,process);
    if(err==noErr) status = false;
    for(int i =0;i< vChannels;i++) {
        free(inBuffers[i]);
    }
    for(int i =0;i< vChannels;i++) {
        free(outBuffers[i]);
    }
    free(inBuffers);
    free(outBuffers);
}
}

bool AudioRender::ConfigureAudioProc(float sampleRate,long bufferSize,int channels,int inChannels,char *device) {
    
    OSStatus err;
    UInt32 size;
    Boolean isWritable;
    
    err = AudioHardwareGetPropertyInfo(kAudioHardwarePropertyDevices,&size,&isWritable);
    if(err!=noErr) return false;
    
    int manyDevices = size/sizeof(AudioDeviceID);
    
    AudioDeviceID devices[manyDevices];
    err = AudioHardwareGetProperty(kAudioHardwarePropertyDevices,&size,&devices);
    if(err!=noErr) return false;
    
    size = sizeof(AudioDeviceID);
    err = AudioHardwareGetProperty(kAudioHardwarePropertyDefaultInputDevice,&size,&vDevice);
    if(err!=noErr) return false;
    
    for(int i=0;i<manyDevices;i++) {
        size = sizeof(char)*256;
        char name[256];
        err = AudioDeviceGetProperty(devices[i],0,false,kAudioDevicePropertyDeviceName,&size,&name);
        if(err!=noErr) return false;
        if(strcmp(device,name)==0) vDevice = devices[i];
    }
    
    char deviceName[256];
    err = AudioDeviceGetProperty(vDevice,0,false,kAudioDevicePropertyDeviceName,&size,&deviceName);
    if(err!=noErr) return false;
    
    printf("DEVICE: %s\n",deviceName);
    
    /*size = sizeof(AudioBufferList);
    AudioBufferList bufList;
    err = AudioDeviceGetProperty(vDevice,0,false,kAudioDevicePropertyStreamConfiguration,&size,&bufList);
    if(err!=noErr) return false;
    
    int nChan = 0;
    for(unsigned int i = 0;i<bufList.mNumberBuffers;i++) {
        nChan += (int)bufList.mBuffers[i].mNumberChannels;
    }*/
    
    
    size = sizeof(AudioStreamBasicDescription);
    AudioStreamBasicDescription SR;
    err = AudioDeviceGetProperty(vDevice,0,false,kAudioDevicePropertyStreamFormat,&size,&SR);
    if(err!=noErr) return false;
    
    err = AudioDeviceGetPropertyInfo(vDevice,0,false,kAudioDevicePropertyStreams,&size,&isWritable);
    if(err!=noErr) return false;
    
    
    vChannels = (int)SR.mChannelsPerFrame*(size/sizeof(AudioStreamID));
    if(vChannels>=channels) vChannels = channels;
    
    printf("OUTPUT CHANNELS: %d\n",vChannels);
    
    err = AudioDeviceGetPropertyInfo(vDevice,0,true,kAudioDevicePropertyStreamFormat,&size,&isWritable);
    if(err!=noErr) { vInChannels = 0; goto endInChan; }
    
    size = sizeof(AudioStreamBasicDescription);
    AudioStreamBasicDescription inSR;
    err = AudioDeviceGetProperty(vDevice,0,true,kAudioDevicePropertyStreamFormat,&size,&inSR);
    if(err!=noErr) return false;
    
    err = AudioDeviceGetPropertyInfo(vDevice,0,true,kAudioDevicePropertyStreams,&size,&isWritable);
    if(err!=noErr) return false;
    
    
    vInChannels = (int)inSR.mChannelsPerFrame*(size/sizeof(AudioStreamID));

endInChan:

    if(vInChannels>=inChannels) vInChannels = inChannels;
    
    printf("INPUT CHANNELS: %d\n",vInChannels);
    
    UInt32 bufFrame;
    size = sizeof(UInt32);
    err = AudioDeviceGetProperty(vDevice,0,false,kAudioDevicePropertyBufferFrameSize,&size,&bufFrame);
    if(err!=noErr) return false;
    
    //bufFrame *= SR.mChannelsPerFrame;
        
    vBufferSize = (long) bufFrame;
    
    if((long)bufFrame!=bufferSize) { 
        printf("I'm trying to set a new buffer size\n");
        UInt32 theSize = sizeof(UInt32);
        UInt32 newBufferSize = (UInt32) bufferSize;///SR.mChannelsPerFrame;
        err = AudioDeviceSetProperty(vDevice,NULL,0,false,kAudioDevicePropertyBufferFrameSize,theSize,&newBufferSize);
        if(err!=noErr) printf("Cannot set a new buffer size\n");
        else {	
            UInt32 newBufFrame;
            size = sizeof(UInt32);
            err = AudioDeviceGetProperty(vDevice,0,false,kAudioDevicePropertyBufferFrameSize,&size,&newBufFrame);
            if(err!=noErr) return false;
            vBufferSize = (long)newBufFrame;//*SR.mChannelsPerFrame;
        }
    }
    
    printf("BUFFER SIZE: %ld\n",vBufferSize);
    
    
    vSampleRate = (float)SR.mSampleRate;
    
    if((float)SR.mSampleRate!=sampleRate) {
        printf("I'm trying to set a new sample rate\n");
        UInt32 theSize = sizeof(AudioStreamBasicDescription);
        SR.mSampleRate = (Float64)sampleRate;
        err = AudioDeviceSetProperty(vDevice,NULL,0,false,kAudioDevicePropertyStreamFormat,theSize,&SR);
        if(err!=noErr) printf("Cannot set a new sample rate\n");
        else {
            size = sizeof(AudioStreamBasicDescription);
            AudioStreamBasicDescription newCheckSR;
            err = AudioDeviceGetProperty(vDevice,0,false,kAudioDevicePropertyStreamFormat,&size,&newCheckSR);
            if(err!=noErr) return false;
            vSampleRate = (float)newCheckSR.mSampleRate;
        }
    }
    
    printf("SAMPLE RATE: %f\n",vSampleRate);
    
    err = AudioDeviceAddIOProc(vDevice,process,this);
    if(err!=noErr) return false;
    
    return true;
}

bool AudioRender::StartAudio() {
if(status) {
    OSStatus err = AudioDeviceStart(vDevice,process);
    if(err!=noErr) return false;
    AudioRender::isProcessing = true;
    return true;
} 
    return false;
}

bool AudioRender::StopAudio() {
if(status) {
    OSStatus err = AudioDeviceStop(vDevice,process);
    if(err!=noErr) return false;
    AudioRender::isProcessing = false;
    return true;
}
    return false;
}

OSStatus AudioRender::process(AudioDeviceID inDevice,const AudioTimeStamp* inNow,const AudioBufferList* inInputData,const AudioTimeStamp* 
                                    inInputTime,AudioBufferList* outOutputData,const AudioTimeStamp* inOutputTime,void* inClientData) 
{
    AudioRender *classe = (AudioRender*)inClientData;
    int channel = 0;
    
    AudioRender::gTime = inInputTime;
    
    for(int i=0;i<AudioRender::gOutputChannels;i++) {
        memset(classe->outBuffers[i],0x0,sizeof(float)*AudioRender::gBufferSize); 
    }
    
if(inInputData->mNumberBuffers!=0 && inInputData->mBuffers[0].mNumberChannels==1) { //MONO MODE PROCESS
    long nframes = inInputData->mBuffers[0].mDataByteSize/sizeof(float);
    
    for(unsigned int a = 0; a < inInputData->mNumberBuffers;a++) {
        memcpy(classe->inBuffers[channel],inInputData->mBuffers[a].mData,inInputData->mBuffers[a].mDataByteSize);
        channel++;
        if(channel==classe->vInChannels) break;
    }
    
    list<ModuleLoader*>::iterator i;
    for(i=ModuleLoader::Audio_StartPoints.begin(); i!=ModuleLoader::Audio_StartPoints.end();++i) {
        ModuleLoader *theConn = *i;
        theConn->audio_executeMethod(nframes);
    }
    
    for(i=ModuleLoader::Audio_ADCs.begin(); i!=ModuleLoader::Audio_ADCs.end();++i) {
        ModuleLoader *theConn = *i;
        theConn->audio_executeMethod(nframes);
    }
     
    goto output;
}

if(inInputData->mNumberBuffers!=0 && inInputData->mBuffers[0].mNumberChannels>1) { //NOT MONO MODE PROCESS
    long nFrames = (inInputData->mBuffers[0].mDataByteSize / sizeof(float))/inInputData->mBuffers[0].mNumberChannels;
    channel = 0;
    
    for(unsigned int b=0;b<inInputData->mNumberBuffers;b++) {
        for(unsigned int a=0; a<inInputData->mBuffers[b].mNumberChannels; a++) {
            for(long i=0;i<nFrames;i++) {
                float *inData = (float*)inInputData->mBuffers[b].mData;
                classe->inBuffers[channel][i] = inData[a+(i*2)];
            }
            channel++;
            if(channel==classe->vInChannels) break;
        }
        if(channel==classe->vInChannels) break;
    }
    
    list<ModuleLoader*>::iterator i;
    for(i=ModuleLoader::Audio_StartPoints.begin(); i!=ModuleLoader::Audio_StartPoints.end();++i) {
        ModuleLoader *theConn = *i;
        theConn->audio_executeMethod(nFrames);
    }
    
    for(i=ModuleLoader::Audio_ADCs.begin(); i!=ModuleLoader::Audio_ADCs.end();++i) {
        ModuleLoader *theConn = *i;
        theConn->audio_executeMethod(nFrames);
    }
    
}

output:

if(outOutputData->mNumberBuffers!=0 && outOutputData->mBuffers[0].mNumberChannels==1) { //MONO MODE PROCESS
    channel = 0;
    
    for(unsigned int a = 0; a < outOutputData->mNumberBuffers;a++) {
        memcpy(outOutputData->mBuffers[a].mData,classe->outBuffers[channel],outOutputData->mBuffers[a].mDataByteSize);
        channel++;
        if(channel==classe->vChannels) break;
    }
    
    goto end;
}

if(outOutputData->mNumberBuffers!=0 && outOutputData->mBuffers[0].mNumberChannels>1) { //NOT MONO MODE PROCESS
    long nFrames = (outOutputData->mBuffers[0].mDataByteSize / sizeof(float))/outOutputData->mBuffers[0].mNumberChannels;
    
    channel = 0;
    
    for(unsigned int b=0;b<outOutputData->mNumberBuffers;b++) {
        for(unsigned int a=0; a<outOutputData->mBuffers[b].mNumberChannels; a++) {
            for(long i=0;i<nFrames;i++) {
                float *outData = (float*)outOutputData->mBuffers[b].mData;
                outData[a+(i*2)] = classe->outBuffers[channel][i];
            }
            channel++;
            if(channel==classe->vChannels) break;
        }
        if(channel==classe->vChannels) break;
    }

}

end:

    return noErr;
}

float **AudioRender::getADC() {
    if(AudioRender::theRender==NULL) return NULL;
    if(AudioRender::isProcessing) return AudioRender::theRender->inBuffers;
    return NULL;
}
float **AudioRender::getDAC() {
    if(AudioRender::theRender==NULL) return NULL;
    if(AudioRender::isProcessing) return AudioRender::theRender->outBuffers;
    return NULL;
}

map<string,AudioUnitRender*> AudioUnitRender::auRenders;
int AudioUnitRender::pluginsModulesK = 0;

AudioUnitRender::AudioUnitRender(char *istanceName,float sampleRate,long bufferSize,int inChannels, int outChannels,void *theAU) 
    : vSampleRate(sampleRate),vBufferSize(bufferSize),vChannels(outChannels),vInChannels(inChannels)
{
    strcpy(&name[0],istanceName);
    theAu = theAU;
    if(vInChannels>0) inBuffers = (float**)malloc(sizeof(float*)*vInChannels);
    else inBuffers = NULL;
    if(vChannels>0) outBuffers = (float**)malloc(sizeof(float*)*vChannels);
    else outBuffers = NULL;
    
    for(int i=0;i<vInChannels;i++) {
        inBuffers[i] = (float*)malloc(sizeof(float)*vBufferSize);
        memset(inBuffers[i],0x0,sizeof(float)*vBufferSize);
    }
    
    for(int i=0;i<vChannels;i++) {
        outBuffers[i] = (float*)malloc(sizeof(float)*vBufferSize);
        memset(outBuffers[i],0x0,sizeof(float)*vBufferSize);
    }
    
    AudioUnitRender::auRenders[istanceName] = this;
    
    AudioRender::gSampleRate = vSampleRate;
    AudioRender::gBufferSize = vBufferSize;
    AudioRender::gInputChannels = vInChannels;
    AudioRender::gOutputChannels = vChannels;
    
}
        
AudioUnitRender::~AudioUnitRender() {

    if(inBuffers) {
        for(int i=0;i<vInChannels;i++) {
            free(inBuffers[i]);
        }
        free(inBuffers);
    }
    
    if(outBuffers) {
        for(int i=0;i<vChannels;i++) {
            free(outBuffers[i]);
        }
        free(outBuffers);
    }
    
}

float **AudioUnitRender::getADC(char *istanceName) {
    map<string,AudioUnitRender*>::iterator i;
    i = AudioUnitRender::auRenders.find(istanceName);
    if(i->second) {
        AudioUnitRender *au = i->second;
        return au->inBuffers;
    } 
    return NULL;
}

float **AudioUnitRender::getDAC(char *istanceName) {
    map<string,AudioUnitRender*>::iterator i;
    i = AudioUnitRender::auRenders.find(istanceName);
    if(i->second) {
        AudioUnitRender *au = i->second;
        return au->outBuffers;
    } 
    return NULL;
}

P_BOOL AudioUnitRender::getAudioConfiguration(char *istanceName,float *outSR,long *outBSize,int *outInCh,int *outOutCh) {
    if(AudioUnitRender::auRenders.size()<=0) return B_NO;
    map<string,AudioUnitRender*>::iterator i;
    i = AudioUnitRender::auRenders.find(istanceName);
    if(i->second) {
        AudioUnitRender *au = i->second;
        *outSR = au->vSampleRate;
        *outBSize = au->vBufferSize;
        *outInCh = au->vInChannels;
        *outOutCh = au->vChannels;
        return B_YES;
    } 
    return B_NO;
}

void AudioUnitRender::addModuleAtInstance(int moduleID,char *instance) {
    map<string,AudioUnitRender*>::iterator i;
    i = AudioUnitRender::auRenders.find(instance);
    if(i->second) {
        AudioUnitRender *au = i->second;
        au->modulesInThis.push_front(moduleID);
    } 
}

P_BOOL AudioUnitRender::closePluginInstance(char *instanceName) {
    map<string,AudioUnitRender*>::iterator i;
    i = AudioUnitRender::auRenders.find(instanceName);
    if(i->second) {
        AudioUnitRender *au = i->second;
        list<int>::iterator it;
        for(it=au->modulesInThis.begin();it!=au->modulesInThis.end();it++) {
            int ID = *it;
            if(ID!=-1) {
                printf("trying to release module: %d\n",ID);
                ModuleLoader *theModule = Connection::moduleFromID(ID);
                if(theModule!=NULL) delete theModule;
            }
        }
        delete au;
        return B_YES;
    }
    return B_NO;
}

int AudioUnitRender::pluginDoRender(char *instance,long nframes) {
    map<string,AudioUnitRender*>::iterator i;
    i = AudioUnitRender::auRenders.find(instance);
    if(i->second) {
        AudioUnitRender *au = i->second;
        list<ModuleLoader*>::iterator i;
        for(i=au->p_StartPoint.begin(); i!=au->p_StartPoint.end();++i) {
            ModuleLoader *theConn = *i;
            theConn->audio_executeMethod(nframes);
        }
    
        for(i=au->p_ADC.begin(); i!=au->p_ADC.end();++i) {
            ModuleLoader *theConn = *i;
            theConn->audio_executeMethod(nframes);
        }
        return TRUE;
    }
    return FALSE;
}

void AudioUnitRender::addStartPoint(char *instance,ModuleLoader *startPoint) {
    map<string,AudioUnitRender*>::iterator i;
    i = AudioUnitRender::auRenders.find(instance);
    if(i->second) {
        AudioUnitRender *au = i->second;
        au->p_StartPoint.push_front(startPoint);
    }
}

void AudioUnitRender::addADC(char *instance,ModuleLoader *adc) {
    map<string,AudioUnitRender*>::iterator i;
    i = AudioUnitRender::auRenders.find(instance);
    if(i->second) {
        AudioUnitRender *au = i->second;
        au->p_ADC.push_front(adc);
    }
}


