/*  

    Panda -  A modular programming language.
    written by Giovanni Petrantoni.
    Copyright (c) 2003, Giovanni Petrantoni.
    http://xpanda.sf.net/
    
    moduleloader.hpp - main Panda classes
    
    Panda is under the terms of the Artistic License :
    
    The Artistic License
    
    Preamble

    The intent of this document is to state the conditions under which a Package may be copied, 
    such that the Copyright Holder maintains some semblance of artistic control over the development of the package, 
    while giving the users of the package the right to use and distribute the Package in a more-or-less customary fashion,
    plus the right to make reasonable modifications.

    Definitions:

    "Package" refers to the collection of files distributed by the Copyright Holder, 
    and derivatives of that collection of files created through textual modification.

    "Standard Version" refers to such a Package if it has not been modified, 
    or has been modified in accordance with the wishes of the Copyright Holder.

    "Copyright Holder" is whoever is named in the copyright or copyrights for the package.

    "You" is you, if you're thinking about copying or distributing this Package.

    "Reasonable copying fee" is whatever you can justify on the basis of media cost, duplication charges, 
    time of people involved, and so on. (You will not be required to justify it to the Copyright Holder, 
    but only to the computing community at large as a market that must bear the fee.)

    "Freely Available" means that no fee is charged for the item itself, 
    though there may be fees involved in handling the item. 
    It also means that recipients of the item may redistribute it under the same conditions they received it.

    1. You may make and give away verbatim copies of the source form of the Standard Version of this Package without restriction, 
    provided that you duplicate all of the original copyright notices and associated disclaimers.

    2. You may apply bug fixes, portability fixes and other modifications derived from the Public Domain or from the Copyright Holder. 
    A Package modified in such a way shall still be considered the Standard Version.

    3. You may otherwise modify your copy of this Package in any way, 
    provided that you insert a prominent notice in each changed file stating how and when you changed that file, 
    and provided that you do at least ONE of the following:

        a) place your modifications in the Public Domain or otherwise make them Freely Available, 
        such as by posting said modifications to Usenet or an equivalent medium, 
        or placing the modifications on a major archive site such as ftp.uu.net, 
        or by allowing the Copyright Holder to include your modifications in the Standard Version of the Package.

        b) use the modified Package only within your corporation or organization.

        c) rename any non-standard executables so the names do not conflict with standard executables, 
        which must also be provided, and provide a separate manual page for each non-standard executable 
        that clearly documents how it differs from the Standard Version.

        d) make other distribution arrangements with the Copyright Holder.

    4. You may distribute the programs of this Package in object code or executable form, 
    provided that you do at least ONE of the following:

    a) distribute a Standard Version of the executables and library files, 
    together with instructions (in the manual page or equivalent) on where to get the Standard Version.

    b) accompany the distribution with the machine-readable source of the Package with your modifications.

    c) accompany any non-standard executables with their corresponding Standard Version executables, 
    giving the non-standard executables non-standard names, and clearly documenting the differences in manual pages (or equivalent), 
    together with instructions on where to get the Standard Version.

    d) make other distribution arrangements with the Copyright Holder.

    5. You may charge a reasonable copying fee for any distribution of this Package. 
    You may charge any fee you choose for support of this Package. You may not charge a fee for this Package itself. 
    However, you may distribute this Package in aggregate with other (possibly commercial) programs as part of 
    a larger (possibly commercial) software distribution provided that you do not advertise this Package as a product of your own.

    6. The scripts and library files supplied as input to or produced as output from the programs of this Package do not automatically 
    fall under the copyright of this Package, but belong to whomever generated them, and may be sold commercially, and may be aggregated with this Package.

    7. C or perl subroutines supplied by you and linked into this Package shall not be considered part of this Package.

    8. The name of the Copyright Holder may not be used to endorse or promote products derived from this software without specific prior written permission.

    9. THIS PACKAGE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, 
    INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.

    The End.

*/


#include <CoreFoundation/CoreFoundation.h>
#include <Carbon/Carbon.h>
#include "panda_h.h"
#include <string.h>
#include <map>
#include <list>
#include <string>
using namespace std;

typedef struct _methodsConnection {
    int idTo;
    int methodTo;
} ConnectedTo;

typedef struct _inletBuffers {
    float **buffers;
    int manyBuffers;
} InletsBuffers;


class Connection;

class AudioRender;

class ModuleLoader {
    public:
        ModuleLoader(const char* name,int moduleId,int argc,char **argv);
        ~ModuleLoader();
        bool IsLoaded() { if(status==1 && module!=NULL) return true; else return false; }
        void executeMethod(int method, Var *valore);
        int audio_executeMethod(long bufferSize);
        void printMethods();
        char * ModuleLoader::getMethodName(int n);
        int ModuleLoader::getMethodKind(int n);
        int getMethodNumber(char *name);
        static list<ModuleLoader*> Audio_StartPoints;
        static list<ModuleLoader*> Audio_ADCs;
        InletsBuffers inletsBufferList;
        float *outletBuffers[256];
        static CFBundleRef GetModuleBundle(int moduleID);
        char *modName;
        Metodi functs;
        MyData *data;
        int getID() { return ID; }
        int theArgc;
        char **theArgv;
    private:
        bool isStartPoint;
        bool isADC;
        CFBundleRef module;
        OSErr CreateBundleFromName(CFStringRef theFrameworkName, CFBundleRef *theBundle);
        void loadBundle(int ID,int argc,char **argv);
        void releaseBundle();
        int getFunctions();
        int status;  
        int ID;
        int outlets,inlets,audio_outlets,audio_inlets;   
        unsigned long timingRef;     
};

class Connection {
    public:
        Connection(int fromID, int toID, int fromOutlet, int toMethod);
        static int executeConnected(int fromID,int fromOutlet,Var *values);
        static void getBuffersConnected(int fromID,int fromOutlet,InletsBuffers2 *outList);
        static int audio_executeConnected(int fromID,int fromOutlet,long bufferSize);
        static void addModule(int ID,ModuleLoader *module);
        static void removeModule(int ID);
        static bool connect(int fromID,int fromOutlet,int toID,int toInlet);
        static bool disconnect(int fromID,int fromOutlet,int toID,int toInlet);
        static bool audio_connect(int fromID,int fromOutlet,int toID,int toInlet);
        static bool audio_disconnect(int fromID,int fromOutlet,int toID,int toInlet);
        static ModuleLoader * Connection::moduleFromID(int ID);
        static map<int,ModuleLoader*>modules;
        static bool savePandaStatusToFile(const char *fileName,int code1,int code2,int code3,int code4);
        static bool loadPandaStatusFromFile(const char *fileName,int code1,int code2,int code3,int code4,loadStatusAudioCallbackPtr call);
        static bool savePandaStatusToFileForPlugin(const char *fileName,int code1,int code2,int code3,int code4);
        static bool loadPandaStatusFromFileForPlugin(const char *fileName,const char *instanceName,int code1,int code2,int code3,int code4,loadStatusAudioCallbackPtr call);
    private:
        int IDfrom,IDto,outletFrom,inletTo;
        static list<Connection*> connections;
        static list<Connection*> audio_connections;
        static map<int,ModuleLoader*>::iterator pos;
};

class AudioRender {
    public:
        AudioRender(float sampleRate,long bufferSize,int inChannels, int outChannels,char *device);
        ~AudioRender();
        bool ConfigureAudioProc(float sampleRate,long bufferSize,int channels,int inChannels, char *device);
        bool StartAudio();
        bool StopAudio();
        static float **getADC();
        static float **getDAC();
        static float gSampleRate;
        static long gBufferSize;
        static int gInputChannels;
        static int gOutputChannels;
        static bool isProcessing;
        static const AudioTimeStamp *gTime;
    private:
        AudioDeviceID vDevice;
        float vSampleRate;
        long vBufferSize;
        int vChannels;//output channels
        int vInChannels; //input chennels
        bool status;
        static OSStatus process(AudioDeviceID inDevice,const AudioTimeStamp* inNow,const AudioBufferList* inInputData,const AudioTimeStamp*  							inInputTime,AudioBufferList* outOutputData,const AudioTimeStamp* inOutputTime,void* inClientData);
        float **inBuffers;
        float **outBuffers;
        static AudioRender *theRender;
};

class AudioUnitRender {
    public:
        AudioUnitRender(char *istanceName,float sampleRate,long bufferSize,int inChannels, int outChannels,void *theAU);
        ~AudioUnitRender();
        static float **getADC(char *istanceName);
        static float **getDAC(char *istanceName);
        static P_BOOL getAudioConfiguration(char *istanceName,float *outSR,long *outBSize,int *outInCh,int *outOutCh);
        static void addModuleAtInstance(int moduleID,char *instance);
        static P_BOOL closePluginInstance(char *instanceName);
        static int pluginDoRender(char *instance,long nframes);
        static void addStartPoint(char *instance,ModuleLoader *startPoint);
        static void addADC(char *instance,ModuleLoader *adc);
        static int pluginsModulesK;
    private:
        list<int> modulesInThis;
        char name[60];
        void *theAu;
        float vSampleRate;
        long vBufferSize;
        int vChannels;//output channels
        int vInChannels; //input chennels
        float **inBuffers;
        float **outBuffers;
        list<ModuleLoader*> p_StartPoint;
        list<ModuleLoader*> p_ADC;
        static map<string,AudioUnitRender*> auRenders;
};


