/*  

    Panda -  A modular programming language.
    written by Giovanni Petrantoni.
    Copyright (c) 2003, Giovanni Petrantoni.
    http://xpanda.sf.net/
    
    pvar.c - Panda variables utilities main code
    
    Panda is under the terms of the Artistic License :
    
    The Artistic License
    
    Preamble

    The intent of this document is to state the conditions under which a Package may be copied, 
    such that the Copyright Holder maintains some semblance of artistic control over the development of the package, 
    while giving the users of the package the right to use and distribute the Package in a more-or-less customary fashion,
    plus the right to make reasonable modifications.

    Definitions:

    "Package" refers to the collection of files distributed by the Copyright Holder, 
    and derivatives of that collection of files created through textual modification.

    "Standard Version" refers to such a Package if it has not been modified, 
    or has been modified in accordance with the wishes of the Copyright Holder.

    "Copyright Holder" is whoever is named in the copyright or copyrights for the package.

    "You" is you, if you're thinking about copying or distributing this Package.

    "Reasonable copying fee" is whatever you can justify on the basis of media cost, duplication charges, 
    time of people involved, and so on. (You will not be required to justify it to the Copyright Holder, 
    but only to the computing community at large as a market that must bear the fee.)

    "Freely Available" means that no fee is charged for the item itself, 
    though there may be fees involved in handling the item. 
    It also means that recipients of the item may redistribute it under the same conditions they received it.

    1. You may make and give away verbatim copies of the source form of the Standard Version of this Package without restriction, 
    provided that you duplicate all of the original copyright notices and associated disclaimers.

    2. You may apply bug fixes, portability fixes and other modifications derived from the Public Domain or from the Copyright Holder. 
    A Package modified in such a way shall still be considered the Standard Version.

    3. You may otherwise modify your copy of this Package in any way, 
    provided that you insert a prominent notice in each changed file stating how and when you changed that file, 
    and provided that you do at least ONE of the following:

        a) place your modifications in the Public Domain or otherwise make them Freely Available, 
        such as by posting said modifications to Usenet or an equivalent medium, 
        or placing the modifications on a major archive site such as ftp.uu.net, 
        or by allowing the Copyright Holder to include your modifications in the Standard Version of the Package.

        b) use the modified Package only within your corporation or organization.

        c) rename any non-standard executables so the names do not conflict with standard executables, 
        which must also be provided, and provide a separate manual page for each non-standard executable 
        that clearly documents how it differs from the Standard Version.

        d) make other distribution arrangements with the Copyright Holder.

    4. You may distribute the programs of this Package in object code or executable form, 
    provided that you do at least ONE of the following:

    a) distribute a Standard Version of the executables and library files, 
    together with instructions (in the manual page or equivalent) on where to get the Standard Version.

    b) accompany the distribution with the machine-readable source of the Package with your modifications.

    c) accompany any non-standard executables with their corresponding Standard Version executables, 
    giving the non-standard executables non-standard names, and clearly documenting the differences in manual pages (or equivalent), 
    together with instructions on where to get the Standard Version.

    d) make other distribution arrangements with the Copyright Holder.

    5. You may charge a reasonable copying fee for any distribution of this Package. 
    You may charge any fee you choose for support of this Package. You may not charge a fee for this Package itself. 
    However, you may distribute this Package in aggregate with other (possibly commercial) programs as part of 
    a larger (possibly commercial) software distribution provided that you do not advertise this Package as a product of your own.

    6. The scripts and library files supplied as input to or produced as output from the programs of this Package do not automatically 
    fall under the copyright of this Package, but belong to whomever generated them, and may be sold commercially, and may be aggregated with this Package.

    7. C or perl subroutines supplied by you and linked into this Package shall not be considered part of this Package.

    8. The name of the Copyright Holder may not be used to endorse or promote products derived from this software without specific prior written permission.

    9. THIS PACKAGE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, 
    INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.

    The End.

*/

#include "pvar.h"
#include <memory.h>
#include <Carbon/Carbon.h>

void * NewVarList(int length) {
    Var *varList = (Var *)malloc(sizeof(Var)*length);
    varList[0].manyVar = length;
    return varList;
}

void DeleteVarList(void *list) {
    Var *varList = (Var*)list;
    if(varList!=NULL) free(varList);
}

void AddToVarListAtIndex(void *list,void *inVar, int index) {
    Var *varList = (Var*)list;
    Var *iVar = (Var*)inVar;
    if(varList[0].manyVar<index) { printf("Array out of bounds"); return; }
    varList[index].var_t = iVar->var_t;
    varList[index].varType = iVar->varType;
    varList[index].length = iVar->length;
    varList[index].specification = iVar->specification;
}

void * varFromListAtIndex(void *list,int index) {
    Var *varList = (Var*)list;
    if(varList[0].manyVar<index) { printf("Array out of bounds"); return NULL; }
    Var *res = &varList[index];
    return res;
}

int manyVars(void *list) {
    Var *varList = (Var*)list;
    return varList->manyVar;
}

void * NewVarArray(int type,int length) {
    switch(type) {
            case INT_T:
            {
                Var *variable;
                variable = (Var*)malloc(sizeof(Var));
                variable->varType = type;
                variable->length = length;
                variable->specification = NULL;
                variable->manyVar = 1;
                variable->var_t = (int*)malloc(sizeof(int)*length);
                return variable;
            }
            case CHAR_T:
            {
                Var *variable;
                variable = (Var*)malloc(sizeof(Var));
                variable->varType = type;
                variable->length = length;
                variable->specification = NULL;
                variable->manyVar = 1;
                variable->var_t = (char*)malloc(sizeof(char)*length);
                return variable;
            }
            case FLOAT_T:
            {
                Var *variable;
                variable = (Var*)malloc(sizeof(Var));
                variable->varType = type;
                variable->length = length;
                variable->specification = NULL;
                variable->manyVar = 1;
                variable->var_t = (float*)malloc(sizeof(float)*length);
                return variable;
            }
            case DOUBLE_T:
            {
                Var *variable;
                variable = (Var*)malloc(sizeof(Var));
                variable->varType = type;
                variable->length = length;
                variable->specification = NULL;
                variable->manyVar = 1;
                variable->var_t = (double*)malloc(sizeof(double)*length);
                return variable;
            }
            case BYTE_T:
            {
                Var *variable;
                variable = (Var*)malloc(sizeof(Var));
                variable->varType = type;
                variable->length = length;
                variable->specification = NULL;
                variable->manyVar = 1;
                variable->var_t = (unsigned char*)malloc(sizeof(unsigned char)*length);
                return variable;
            }
            case ULONG_T:
            {
                Var *variable;
                variable = (Var*)malloc(sizeof(Var));
                variable->varType = type;
                variable->length = length;
                variable->specification = NULL;
                variable->manyVar = 1;
                variable->var_t = (unsigned long*)malloc(sizeof(unsigned long)*length);
                return variable;
            }
            case LONG_T:
            {
                Var *variable;
                variable = (Var*)malloc(sizeof(Var));
                variable->varType = type;
                variable->length = length;
                variable->specification = NULL;
                variable->manyVar = 1;
                variable->var_t = (long*)malloc(sizeof(long)*length);
                return variable;
            }
            case LDOUBLE_T:
            {
                Var *variable;
                variable = (Var*)malloc(sizeof(Var));
                variable->varType = type;
                variable->length = length;
                variable->specification = NULL;
                variable->manyVar = 1;
                variable->var_t = (long double*)malloc(sizeof(long double)*length);
                return variable;
            }
            case OBJECT_T:
            {
                Var *variable;
                variable = (Var*)malloc(sizeof(Var));
                variable->varType = type;
                variable->length = length;
                variable->specification = NULL;
                variable->var_t = NULL;
                variable->manyVar = 1;
                return variable;
            }
            case PTR_T:
            {
                Var *variable;
                variable = (Var*)malloc(sizeof(Var));
                variable->varType = type;
                variable->length = length;
                variable->specification = NULL;
                variable->var_t = NULL;
                variable->manyVar = 1;
                return variable;
            }
            case BANG_T:
            {
                Var *variable;
                variable = (Var*)malloc(sizeof(Var));
                variable->varType = type;
                variable->length = length;
                variable->specification = NULL;
                variable->var_t = (char*)malloc(sizeof(char));
                variable->manyVar = 1;
                return variable;
            }
            default:
            {
                printf("Error: Type not recoginzed\n");
                return NULL;
            }
    }
    return NULL;
}

void * NewVar(int type) {
    return NewVarArray(type,1);
}

void DeleteVar(void *varP) {
    Var *var = (Var*)varP;
    if(var!=NULL) {
            if(var->var_t!=NULL) { free(var->var_t); var->var_t = NULL; }
            if(var->specification!=NULL) { free(var->specification); var->specification = NULL; }
            free(var);
        var = NULL;
    }
}

int type(void *varP) {
    Var *var = (Var*)varP;
    return var->varType;
}

int length(void *varP) {
    Var *var = (Var*)varP;
    return var->length;
}

char *specifications(void *varP) {
    Var *var = (Var*)varP;
    return var->specification;
}

void setSpecifications(void *varP,char *specs) {
    Var *var = (Var*)varP;
    if(var->specification!=NULL) free(var->specification);
    var->specification = (char*)malloc(sizeof(char)*256);
    strcpy(var->specification,specs);
}

void setIntValue(void *varP,int value) {
    Var *var = (Var*)varP;
    if(var->varType!=INT_T) { printf("This variable is not an INT\n"); return; }
    int *arr = (int*)var->var_t;
    *arr = value;
}

void setIntValueAtIndex(void *varP,int value,int index) {
    Var *var = (Var*)varP;
    if(var->varType!=INT_T) { printf("This variable is not an INT\n"); return; }
    if(var->length<index) { printf("Array out of bounds\n"); return; }
    int *arr = (int*) var->var_t;
    arr[index] = value;
}

void setCharValue(void *varP,char value) {
    Var *var = (Var*)varP;
    if(var->varType!=CHAR_T) { printf("This variable is not a CHAR\n"); return; }
    char *arr = (char*)var->var_t;
    *arr = value;
}

void setCharValueAtIndex(void *varP,char value,int index) {
    Var *var = (Var*)varP;
    if(var->varType!=CHAR_T) { printf("This variable is not a CHAR\n"); return; }
    if(var->length<index) { printf("Array out of bounds\n"); return; }
    char *arr = (char*)var->var_t;
    arr[index] = value;
}

void setStringValue(void *varP,char *value) {
    Var *var = (Var*)varP;
    if(var->varType!=CHAR_T) { printf("This variable is not a CHAR\n"); return; }
    if(var->length<strlen(value)) { printf("Array out of bounds\n"); return; }
    char *arr = (char*)var->var_t;
    strcpy(arr,value);
}

void setFloatValue(void *varP,float value) {
    Var *var = (Var*)varP;
    if(var->varType!=FLOAT_T) { printf("This variable is not a FLOAT\n"); return; }
    float *arr = (float*)var->var_t;
    *arr = value;
}

void setFloatValueAtIndex(void *varP,float value,int index) {
    Var *var = (Var*)varP;
    if(var->varType!=FLOAT_T) { printf("This variable is not a FLOAT\n"); return; }
    if(var->length<index) { printf("Array out of bounds\n"); return; }
    float *arr = (float*)var->var_t;
    arr[index] = value;
}

void setDoubleValue(void *varP,double value) {
    Var *var = (Var*)varP;
    if(var->varType!=DOUBLE_T) { printf("This variable is not a DOUBLE\n"); return; }
    double *arr = (double*) var->var_t;
    *arr = value;
}

void setDoubleValueAtIndex(void *varP,double value,int index) {
    Var *var = (Var*)varP;
    if(var->varType!=DOUBLE_T) { printf("This variable is not a DOUBLE\n"); return; }
    if(var->length<index) { printf("Array out of bounds\n"); return; }
    double *arr = (double*) var->var_t;
    arr[index] = value;
}

void setByteValue(void *varP,unsigned char value) {
    Var *var = (Var*)varP;
    if(var->varType!=BYTE_T) { printf("This variable is not a unsigned char\n"); return; }
    unsigned char *arr = (unsigned char*)var->var_t;
    *arr = value;
}

void setByteValueAtIndex(void *varP,unsigned char value,int index) {
    Var *var = (Var*)varP;
    if(var->varType!=BYTE_T) { printf("This variable is not a unsigned char\n"); return; }
    if(var->length<index) { printf("Array out of bounds\n"); return; }
    unsigned char *arr = (unsigned char*)var->var_t;
    arr[index] = value;
}

void setULongValue(void *varP,unsigned long value) {
    Var *var = (Var*)varP;
    if(var->varType!=ULONG_T) { printf("This variable is not an UNSIGNED LONG\n"); return; }
    unsigned long *arr = (unsigned long*)var->var_t;
    *arr = value;
}

void setULongValueAtIndex(void *varP,unsigned long value,int index) {
    Var *var = (Var*)varP;
    if(var->varType!=ULONG_T) { printf("This variable is not an UNSIGNED LONG\n"); return; }
    if(var->length<index) { printf("Array out of bounds\n"); return; }
    unsigned long *arr = (unsigned long*)var->var_t;
    arr[index] = value;
}

void setLongValue(void *varP,long value) {
    Var *var = (Var*)varP;
    if(var->varType!=LONG_T) { printf("This variable is not an UNSIGNED INT\n"); return; }
    long *arr = (long*)var->var_t;
    *arr = value;
}

void setLongValueAtIndex(void *varP,long value,int index) {
    Var *var = (Var*)varP;
    if(var->varType!=LONG_T) { printf("This variable is not an UNSIGNED INT\n"); return; }
    if(var->length<index) { printf("Array out of bounds\n"); return; }
    long *arr = (long*)var->var_t;
    arr[index] = value;
}

void setLDoubleValue(void *varP,long double value) {
    Var *var = (Var*)varP;
    if(var->varType!=LDOUBLE_T) { printf("This variable is not a LONG DOUBLE\n"); return; }
    long double *arr = (long double*)var->var_t;
    *arr = value;
}

void setLDoubleValueAtIndex(void *varP,long double value,int index) {
    Var *var = (Var*)varP;
    if(var->varType!=LDOUBLE_T) { printf("This variable is not a LONG DOUBLE\n"); return; }
    if(var->length<index) { printf("Array out of bounds\n"); return; }
    long double *arr = (long double*)var->var_t;
    arr[index] = value;
}

void setObjectValue(void *varP,void *value) {
    Var *var = (Var*)varP;
    if(var->varType!=OBJECT_T) { printf("This variable is not an OBJECT\n"); return; }
    var->var_t = value;
}


void setPtrValue(void *varP,void *value) {
    Var *var = (Var*)varP;
    if(var->varType!=PTR_T) { printf("This variable is not a PTR\n"); return; }
    var->var_t = value;
}

int intValue(void *varP) {
    Var *var = (Var*)varP;
    if(var->varType!=INT_T) { printf("This variable is not an INT\n"); return NULL; }
    int *res = (int*)var->var_t;
    return *res;
}

int intValueAtIndex(void *varP,int index) {
    Var *var = (Var*)varP;
    if(var->varType!=INT_T) { printf("This variable is not an INT\n"); return NULL; }
    if(var->length<index) { printf("Array out of bounds\n"); return NULL; }
    int *res = (int*)var->var_t;
    return res[index];
}

char charValue(void *varP) {
    Var *var = (Var*)varP;
    if(var->varType!=CHAR_T) { printf("This variable is not a CHAR\n"); return NULL; }
    char *res = (char*)var->var_t;
    return *res;
}

char charValueAtIndex(void *varP,int index) {
    Var *var = (Var*)varP;
    if(var->varType!=CHAR_T) { printf("This variable is not an CHAR\n"); return NULL; }
    if(var->length<index) { printf("Array out of bounds\n"); return NULL; }
    char *res = (char*)var->var_t;
    return res[index];
}

char * stringValue(void *varP) {
    Var *var = (Var*)varP;
    if(var->varType!=CHAR_T) { printf("This variable is not an CHAR\n"); return NULL; }
    char *res = (char*)var->var_t;
    return res;
}


float floatValue(void *varP) {
    Var *var = (Var*)varP;
    if(var->varType!=FLOAT_T) { printf("This variable is not a FLOAT\n"); return NULL; }
    float *res = (float*)var->var_t;
    return *res;
}

float floatValueAtIndex(void *varP,int index) {
    Var *var = (Var*)varP;
    if(var->varType!=FLOAT_T) { printf("This variable is not a FLOAT\n"); return NULL; }
    if(var->length<index) { printf("Array out of bounds\n"); return NULL; }
    float *res = (float*)var->var_t;
    return res[index];
}

double doubleValue(void *varP) {
    Var *var = (Var*)varP;
    if(var->varType!=DOUBLE_T) { printf("This variable is not a DOUBLE\n"); return NULL; }
    double *res = (double*)var->var_t;
    return *res;
}

double doubleValueAtIndex(void *varP,int index) {
    Var *var = (Var*)varP;
    if(var->varType!=DOUBLE_T) { printf("This variable is not a DOUBLE\n"); return NULL; }
    if(var->length<index) { printf("Array out of bounds\n"); return NULL; }
    double *res = (double*)var->var_t;
    return res[index];
}

unsigned char byteValue(void *varP) {
    Var *var = (Var*)varP;
    if(var->varType!=BYTE_T) { printf("This variable is not a unsigned char\n"); return NULL; }
    unsigned char *res = (unsigned char*)var->var_t;
    return *res;
}

unsigned char byteValueAtIndex(void *varP,int index) {
    Var *var = (Var*)varP;
    if(var->varType!=BYTE_T) { printf("This variable is not a unsigned char\n"); return NULL; }
    if(var->length<index) { printf("Array out of bounds\n"); return NULL; }
    unsigned char *res = (unsigned char*)var->var_t;
    return res[index];
}

unsigned long uLongValue(void *varP) {
    Var *var = (Var*)varP;
    if(var->varType!=ULONG_T) { printf("This variable is not an UNSIGNED LONG\n"); return NULL; }
    unsigned long *res = (unsigned long*)var->var_t;
    return *res;
}

unsigned long uLongValueAtIndex(void *varP,int index) {
    Var *var = (Var*)varP;
    if(var->varType!=ULONG_T) { printf("This variable is not an UNSIGNED LONG\n"); return NULL; }
    if(var->length<index) { printf("Array out of bounds\n"); return NULL; }
    unsigned long *res = (unsigned long*)var->var_t;
    return res[index];
}

long longValue(void *varP) {
    Var *var = (Var*)varP;
    if(var->varType!=LONG_T) { printf("This variable is not a LONG\n"); return NULL; }
    long *res = (long*)var->var_t;
    return *res;
}

long longValueAtIndex(void *varP,int index) {
    Var *var = (Var*)varP;
    if(var->varType!=LONG_T) { printf("This variable is not a LONG\n"); return NULL; }
    if(var->length<index) { printf("Array out of bounds\n"); return NULL; }
    long *res = (long*)var->var_t;
    return res[index];
}

long double lDoubleValue(void *varP) {
    Var *var = (Var*)varP;
    if(var->varType!=LDOUBLE_T) { printf("This variable is not a LDOUBLE\n"); return NULL; }
    long double *res = (long double*)var->var_t;
    return *res;
}

long double lDoubleValueAtIndex(void *varP,int index) {
    Var *var = (Var*)varP;
    if(var->varType!=LDOUBLE_T) { printf("This variable is not a LONG DOUBLE\n"); return NULL; }
    if(var->length<index) { printf("Array out of bounds\n"); return NULL; }
    long double *res = (long double*)var->var_t;
    return res[index];
}

void * objectValue(void *varP) {
    Var *var = (Var*)varP;
    if(var->varType!=OBJECT_T) { printf("This variable is not an OBJECT\n"); return NULL; }
    return var->var_t;
}

void * ptrValue(void *varP) {
    Var *var = (Var*)varP;
    if(var->varType!=PTR_T) { printf("This variable is not a PTR\n"); return NULL; }
    return var->var_t;
}
