/*  

    Panda -  A modular programming language.
    written by Giovanni Petrantoni.
    Copyright (c) 2003, Giovanni Petrantoni.
    http://xpanda.sf.net/
    
    panda_h.h - main internal header
    
    Panda is under the terms of the Artistic License :
    
    The Artistic License
    
    Preamble

    The intent of this document is to state the conditions under which a Package may be copied, 
    such that the Copyright Holder maintains some semblance of artistic control over the development of the package, 
    while giving the users of the package the right to use and distribute the Package in a more-or-less customary fashion,
    plus the right to make reasonable modifications.

    Definitions:

    "Package" refers to the collection of files distributed by the Copyright Holder, 
    and derivatives of that collection of files created through textual modification.

    "Standard Version" refers to such a Package if it has not been modified, 
    or has been modified in accordance with the wishes of the Copyright Holder.

    "Copyright Holder" is whoever is named in the copyright or copyrights for the package.

    "You" is you, if you're thinking about copying or distributing this Package.

    "Reasonable copying fee" is whatever you can justify on the basis of media cost, duplication charges, 
    time of people involved, and so on. (You will not be required to justify it to the Copyright Holder, 
    but only to the computing community at large as a market that must bear the fee.)

    "Freely Available" means that no fee is charged for the item itself, 
    though there may be fees involved in handling the item. 
    It also means that recipients of the item may redistribute it under the same conditions they received it.

    1. You may make and give away verbatim copies of the source form of the Standard Version of this Package without restriction, 
    provided that you duplicate all of the original copyright notices and associated disclaimers.

    2. You may apply bug fixes, portability fixes and other modifications derived from the Public Domain or from the Copyright Holder. 
    A Package modified in such a way shall still be considered the Standard Version.

    3. You may otherwise modify your copy of this Package in any way, 
    provided that you insert a prominent notice in each changed file stating how and when you changed that file, 
    and provided that you do at least ONE of the following:

        a) place your modifications in the Public Domain or otherwise make them Freely Available, 
        such as by posting said modifications to Usenet or an equivalent medium, 
        or placing the modifications on a major archive site such as ftp.uu.net, 
        or by allowing the Copyright Holder to include your modifications in the Standard Version of the Package.

        b) use the modified Package only within your corporation or organization.

        c) rename any non-standard executables so the names do not conflict with standard executables, 
        which must also be provided, and provide a separate manual page for each non-standard executable 
        that clearly documents how it differs from the Standard Version.

        d) make other distribution arrangements with the Copyright Holder.

    4. You may distribute the programs of this Package in object code or executable form, 
    provided that you do at least ONE of the following:

    a) distribute a Standard Version of the executables and library files, 
    together with instructions (in the manual page or equivalent) on where to get the Standard Version.

    b) accompany the distribution with the machine-readable source of the Package with your modifications.

    c) accompany any non-standard executables with their corresponding Standard Version executables, 
    giving the non-standard executables non-standard names, and clearly documenting the differences in manual pages (or equivalent), 
    together with instructions on where to get the Standard Version.

    d) make other distribution arrangements with the Copyright Holder.

    5. You may charge a reasonable copying fee for any distribution of this Package. 
    You may charge any fee you choose for support of this Package. You may not charge a fee for this Package itself. 
    However, you may distribute this Package in aggregate with other (possibly commercial) programs as part of 
    a larger (possibly commercial) software distribution provided that you do not advertise this Package as a product of your own.

    6. The scripts and library files supplied as input to or produced as output from the programs of this Package do not automatically 
    fall under the copyright of this Package, but belong to whomever generated them, and may be sold commercially, and may be aggregated with this Package.

    7. C or perl subroutines supplied by you and linked into this Package shall not be considered part of this Package.

    8. The name of the Copyright Holder may not be used to endorse or promote products derived from this software without specific prior written permission.

    9. THIS PACKAGE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, 
    INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.

    The End.

*/

#include <stdio.h>
#include <Carbon/Carbon.h>
#include <CoreAudio/CoreAudio.h>
#define DEBUG 1

enum { //Variable types, (varType);
    INT_T,CHAR_T,FLOAT_T,DOUBLE_T,BYTE_T,ULONG_T,LONG_T,LDOUBLE_T,OBJECT_T,PTR_T,BANG_T
};

enum {
    B_YES = 1, B_NO = 2
};

enum { //Module types, (TYPE);
    AUDIO_PARAMETERS = 1, ONLY_PARAMETERS = 3
};

typedef unsigned char P_BOOL;
typedef unsigned char P_BYTE;

typedef struct _inletBuf {
    float *buffers[256];
    int manyBuffers;
} InletsBuffers2;


typedef struct _variable {
    int varType;
    void *var_t;
    int manyVar;
    int length;
    char *specification;
} Var;

typedef struct _sharedPtr {
    char key[256];
    void *ptr;
} S_SharedPtr;

typedef int TYPE;

typedef void MyData;

typedef struct _parametersList {
    TYPE moduleType;
    int manyMethods;
    int outlets;
    int inlets;
    int moduleID;
    int audio_outlets; //olny if AUDIO type is declared, 1 outlet is 1 audio output channel
    int audio_inlets; //olny if AUDIO type is declared, 1 inlet is 1 audio input channel
    P_BOOL audio_startpoint; //if your AUDIO module has no audio input inlets it will need an audio_startpoint
}ParamList;

typedef struct metodi { 
    MyData * (*NewPandaMod)(int *err,int MODULE_ID,int argc,char **argv);
    void (*InletCallBackPtr) (MyData *x,int inlet, Var *inValues);
    float ** (*AllocInletsPtr) (MyData *x,long bufferSize, int manyBuffers);
    int (*AudioProcessPtr)(MyData *x,long nframes);
    ParamList (*GetParameterListPtr)(MyData *x);
    void (*flushModulePtr)(MyData *x);
}Metodi;

typedef void (*loadStatusAudioCallbackPtr)(bool status);

#ifdef __cplusplus
extern "C" {
#endif
void outletCall(int ID, int outletNumber,void * values);
int audio_StartPointCall(int ID, int outletNumber,long bufferSize);
InletsBuffers2 GetOutletBuffers(int ID, int outletNumber);
float **GetADCBuffers(int *channels);
float **GetDACBuffers(int *channels);
long GetBufferSize();
float GetSampleRate();
int GetDeviceInputChannels();
int GetDeviceOutputChannels();
const AudioTimeStamp * GetDeviceTime();
CFBundleRef GetModuleBundleFromID(int ID);
void PandaLog(char *fmt,...);
int setSharedPointer(char *key,void *ptr); 
void *sharedPointer(char *key); 
void removeSharedPointer(char *key);
float **GetPluginADCBuffers(char *istanceName);
float **GetPluginDACBuffers(char *istanceName);
P_BOOL GetPluginAudioConfiguration(char *istanceName,float *outSR,long *outBSize,int *outInCh,int *outOutCh);

//UTILITIES
Var *initNewVar(int *types,int many,void **values);
int c_bool(P_BOOL the_p_bool);
P_BOOL p_bool(int c_cond);


#ifdef __cplusplus
}
#endif
